@@ -1,7 +1,9 @@
 import hashlib
 import os.path
 import re
-from typing import Any, Iterable, Mapping, Optional
+from typing import Any, Iterable, Mapping, Optional, Type, Union
+
+import pandas as pd
 
 PARAM_DIGEST_HASH = "sha256"
 
@@ -11,6 +13,8 @@ class BaseParam:
 
     label: str = ""
     value: Any = None
+    hide: bool = False
+    read_only: bool = False
 
     def copy(self):
         """Plugins declare their parameters with instances of BaseParam, these
@@ -38,7 +42,6 @@ class SimpleParam(BaseParam):
     """A SimpleParam has a single value"""
 
     var_type: type = type(None)
-    read_only: bool = False
 
     def __init__(self, label: str, value=None, read_only: bool = False):
         self.label = label
@@ -48,7 +51,10 @@ class SimpleParam(BaseParam):
             self.read_only = True
 
     def clean_value(self, value):
-        return self.var_type(value)
+        try:
+            return self.var_type(value)
+        except ValueError:
+            return None
 
     @property
     def value(self):
@@ -62,7 +68,7 @@ class SimpleParam(BaseParam):
     def value(self):
         self._value = None
 
-    def copy(self):
+    def copy(self) -> "SimpleParam":
         return self.__class__(self.label, self.value, self.read_only)
 
 
@@ -84,11 +90,28 @@ class IntegerParam(SimpleParam):
     var_type = int
     _value: int = 0
 
+    def clean_value(self, value):
+        if isinstance(value, str):
+            return super().clean_value("".join(re.split(r"\D+", value)))
+        else:
+            return super().clean_value(value)
+
 
 class FloatParam(SimpleParam):
     var_type = float
     _value: float = 0.0
 
+    def clean_value(self, value):
+        if isinstance(value, str):
+            try:
+                a, b = value.split(".", 1)
+                s = re.split(r"\D+", a) + ["."] + re.split(r"\D+", b)
+            except ValueError:
+                s = re.split(r"\D+", value)
+            return super().clean_value("".join(s))
+        else:
+            return super().clean_value(value)
+
 
 class StringParam(SimpleParam):
     var_type = str
@@ -131,10 +154,8 @@ class StringCharacterSetParam(StringParam):
         x = "".join([self.clean_character(c) for c in value_str])
         return x
 
-    def copy(self):
-        return self.__class__(
-            self.label, self.value, self.read_only, character_set=self.character_set
-        )
+    def copy(self) -> "StringCharacterSetParam":
+        return self.__class__(self.label, self.value, self.read_only, character_set=self.character_set)
 
 
 class FileParam(StringParam):
@@ -170,9 +191,17 @@ class FileParam(StringParam):
             return "0"
 
     def get_parameters(self, key, base_dir="."):
-        return ((key, os.path.relpath(self.value, base_dir)),)
+        if self.value:
+            if base_dir:
+                path = os.path.relpath(self.value, base_dir)
+            else:
+                path = os.path.abspath(self.value)
+        else:
+            path = None
 
-    def copy(self):
+        return [(key, path)]
+
+    def copy(self) -> "FileParam":
         return self.__class__(self.label, self.value, self.read_only, file_types=self.file_types)
 
     def get_hash_value(self) -> str:
@@ -191,7 +220,7 @@ class FileSaveParam(StringParam):
         if file_types is not None:
             self.file_types = file_types
 
-    def clean_value(self, value: str | tuple | list, file_types=None):
+    def clean_value(self, value: Union[str, tuple, list], file_types=None):
         try:
             return os.path.relpath(str(value))
         except ValueError:
@@ -202,7 +231,11 @@ class ChoiceParam(BaseParam):
     """A drop-down menu parameter choosing between options.
     Defaults to 'None'"""
 
+    DEFAULT_VALUE = ""
+
     _value: Optional[str] = None
+    _choice: Optional[int] = None
+    choices: list[str] = []
 
     def __init__(
         self,
@@ -211,7 +244,7 @@ class ChoiceParam(BaseParam):
         choices: Optional[Iterable[str]] = None,
     ):
         self.label = label
-        self._value = value
+        self.value = value
         self.choices = list(choices or [])
 
     @property
@@ -220,33 +253,49 @@ class ChoiceParam(BaseParam):
 
     @value.setter
     def value(self, value):
-        if value in self.choices:
-            self._value = value
+        if value is None:
+            self._value = self.DEFAULT_VALUE
         else:
+            self._value = value
+        try:
+            self._choice = self.choices.index(value)
+        except ValueError:
+            self._choice = None
+
+    @property
+    def choice(self):
+        return self._choice
+
+    @choice.setter
+    def choice(self, choice):
+        if choice is not None and 0 <= choice < len(self.choices):
+            self._choice = choice
+            self._value = self.choices[choice]
+        else:
+            self._choice = None
             self._value = None
 
     def set_choices(self, choices: Iterable[str]):
         self.choices = list(choices)
-        if len(self.choices) == 1:
-            self._value = list(choices)[0]
-        elif self.value not in self.choices:
-            self._value = None
+        if self.choices:
+            if self._value not in self.choices:
+                self._value = self.choices[0]
+        else:
+            self._value = self.DEFAULT_VALUE
 
-    def copy(self):
+    def copy(self) -> "ChoiceParam":
         return self.__class__(self.label, self.value, self.choices)
 
 
 class DataTypeChoiceParam(ChoiceParam):
-    DATA_TYPES: Mapping[str, Optional[tuple[type, Any]]] = {
-        "string": (str, None),
-        "number": (float, None),
-        "integer": (int, 0),
-        "boolean": (bool, None),
+    DATA_TYPES: Mapping[str, tuple[type, Any, Type[SimpleParam]]] = {
+        "string": (str, "", StringParam),
+        "number": (float, 0.0, FloatParam),
+        "integer": (int, 0, IntegerParam),
+        "boolean": (bool, False, BooleanParam),
     }
 
-    def __init__(
-        self, label: str, value: Optional[str] = None, choices: Optional[Iterable[str]] = None
-    ):
+    def __init__(self, label: str, value: Optional[str] = None, choices: Optional[Iterable[str]] = None):
         if not choices:
             choices = list(self.DATA_TYPES.keys())
         super().__init__(label, value, choices)
@@ -258,48 +307,51 @@ class DataTypeChoiceParam(ChoiceParam):
             return self.DATA_TYPES[self.value][0]
 
     def cast_value(self, value):
-        if self.value is None:
-            return None
         if value is None:
             return self.DATA_TYPES[self.value][1]
         else:
             return self.DATA_TYPES[self.value][0](value)
 
+    def get_parameter(self, label: str, value=None) -> BaseParam:
+        return self.DATA_TYPES[self.value][2](label, value)
+
 
 class DataTypeOrNoneChoiceParam(DataTypeChoiceParam):
-    NONE_VALUE = "— NONE —"
+    DEFAULT_VALUE = "— NONE —"
 
-    DATA_TYPES = {
-        "string": (str, None),
-        "number": (float, None),
-        "integer": (int, 0),
-        "boolean": (bool, None),
-        NONE_VALUE: None,
-    }
-
-    def __init__(
-        self, label: str, value: Optional[str] = None, choices: Optional[Iterable[str]] = None
-    ):
+    def __init__(self, label: str, value: Optional[str] = None, choices: Optional[Iterable[str]] = None):
         if not choices:
-            choices = list(self.DATA_TYPES.keys())
-        if value is None:
-            value = self.NONE_VALUE
+            choices = list(self.DATA_TYPES.keys()) + [self.DEFAULT_VALUE]
         super().__init__(label, value, choices)
 
     def get_selected_type(self):
-        if self.value == self.NONE_VALUE:
+        if self.value == self.DEFAULT_VALUE:
             return None
         else:
             return super().get_selected_type()
 
     def cast_value(self, value):
-        if self.value == self.NONE_VALUE:
+        if self.value == self.DEFAULT_VALUE:
             return None
         else:
             return super().cast_value(value)
 
     def is_none(self):
-        return self.value == self.NONE_VALUE
+        return self.value == self.DEFAULT_VALUE
+
+    def is_not_none(self):
+        return self.value != self.DEFAULT_VALUE
+
+
+def _dataframe_get_column(df: pd.DataFrame, col: str):
+    if col in df.columns:
+        return df[col]
+    elif col == df.index.name:
+        return df.index.to_series()
+    elif hasattr(df.index, "names") and col in df.index.names:
+        return df.index.to_frame()[col]
+    else:
+        raise ValueError(f"Column {col} not found")
 
 
 class ColumnChoiceParam(ChoiceParam):
@@ -309,24 +361,71 @@ class ColumnChoiceParam(ChoiceParam):
     def set_column_choices(self, choices):
         self.set_choices(list(choices))
 
-class ColumnOrIndexChoiceParam(ColumnChoiceParam):
-    INDEX_VALUE = "— INDEX —"
-
-    def set_choices(self, choices: Iterable[str]):
-        super().set_choices([self.INDEX_VALUE] + list(choices))
-
-    def is_index(self):
-        return self.value == self.INDEX_VALUE
+    def get_column(self, df):
+        return _dataframe_get_column(df, self.value)
 
 
 class ColumnOrNoneChoiceParam(ColumnChoiceParam):
-    NONE_VALUE = "— NONE —"
+    DEFAULT_VALUE = "— NONE —"
 
     def set_choices(self, choices: Iterable[str]):
-        super().set_choices([self.NONE_VALUE] + list(choices))
+        super().set_choices([self.DEFAULT_VALUE] + list(choices))
 
     def is_none(self):
-        return self.value == self.NONE_VALUE
+        return self.value == self.DEFAULT_VALUE
+
+    def is_not_none(self):
+        return self.value != self.DEFAULT_VALUE
+
+    def get_column(self, df):
+        if self.value == self.DEFAULT_VALUE:
+            return None
+        else:
+            return _dataframe_get_column(df, self.value)
+
+
+class ColumnOrIndexChoiceParam(ColumnChoiceParam):
+    DEFAULT_VALUE = "— INDEX —"
+
+    def set_choices(self, choices: Iterable[str]):
+        super().set_choices([self.DEFAULT_VALUE] + list(choices))
+
+    def is_index(self):
+        return self.value == self.DEFAULT_VALUE
+
+    def is_not_index(self):
+        return self.value != self.DEFAULT_VALUE
+
+    def get_column(self, df):
+        if self.value == self.DEFAULT_VALUE:
+            return df.index.to_series()
+        else:
+            return _dataframe_get_column(df, self.value)
+
+
+class ColumnOrStringParam(ColumnChoiceParam):
+    DEFAULT_VALUE = ""
+    PREFIX = "— "
+
+    def set_column_choices(self, choices):
+        self.set_choices([self.PREFIX + c for c in choices])
+
+    def get_column_name(self):
+        if self.value.startswith(self.PREFIX):
+            return self.value[len(self.PREFIX) :]
+        return None
+
+    def get_value(self, data: dict):
+        if self.value.startswith(self.PREFIX):
+            return data[self.value[len(self.PREFIX) :]]
+        else:
+            return self.value
+
+    def set_choices(self, choices: Iterable[str]):
+        self.choices = list(choices)
+        if self._value is not None and self._value.startswith(self.PREFIX) and self._value not in self.choices:
+            self._value = self.DEFAULT_VALUE
+            self._choice = None
 
 
 class ArrayParam(BaseParam):
@@ -337,7 +436,7 @@ class ArrayParam(BaseParam):
     # maybe we should have a TabularParam which combines ArrayParam and
     # MultiParam more directly."""
 
-    params: list[BaseParam] = []
+    params: list[BaseParam]
 
     def __init__(
         self,
@@ -379,7 +478,8 @@ class ArrayParam(BaseParam):
 
     def relabel(self):
         for n, param in enumerate(self.params):
-            param.label = self.param.label + f" {n+1}"
+            if param.label.startswith(self.param.label + " "):
+                param.label = self.param.label + f" {n+1}"
 
     def copy(self) -> "ArrayParam":
         return self.__class__(self.label, self.param, self.read_only, self.min_size, self.max_size)
@@ -432,6 +532,37 @@ class ArrayParam(BaseParam):
         for p in self.params:
             p.set_column_choices(choices)
 
+
+class PerColumnArrayParam(ArrayParam):
+    """An ArrayParam where each value in the array corresponds to a column
+    in the input dataframe, as set by set_column_choices."""
+
+    def __init__(self, *a, **k) -> None:
+        super().__init__(*a, **k)
+        self.read_only = True
+
+    def get_parameters(self, key, base_dir="."):
+        for n, p in enumerate(self.params):
+            yield f"{key}.{n}._label", p.label
+            yield from p.get_parameters(f"{key}.{n}", base_dir)
+
+    def set_column_choices(self, choices):
+        params_by_label = {p.label: p for p in self.params}
+        self.params = [None] * len(choices)
+        for num, name in enumerate(choices):
+            label = f'"{name}"'
+            if label in params_by_label:
+                self.params[num] = params_by_label[label]
+            else:
+                self.params[num] = self.param.copy()
+                self.params[num].label = label
+            self.params[num].set_column_choices(choices)
+
+    def get_column_params(self):
+        for p in self.params:
+            yield p.label, p
+
+
 class FileArrayParam(ArrayParam):
     """FileArrayParam is an ArrayParam arranged per-file.  Using this class
     really just marks it as expecting to be populated from an open file
@@ -515,9 +646,14 @@ class MultiParam(BaseParam):
     def get_hash_value(self):
         digest = hashlib.new(PARAM_DIGEST_HASH)
         for k, p in self.params.items():
-            digest.update((k + "\0" + p.get_hash_value()).encode("utf-8"))
+            digest.update((str(k) + "\0" + p.get_hash_value()).encode("utf-8"))
         return digest.hexdigest()
 
     def set_column_choices(self, choices):
         for p in self.params.values():
             p.set_column_choices(choices)
+
+
+class TabularMultiParam(MultiParam):
+    """This is just used to drop a hint to the GUI as to how the MultiParam
+    is to be presented ... as a hierarchy or as a table ..."""
