@@ -1,11 +1,12 @@
 import csv
-from io import StringIO
-from typing import Any, Optional
+import gzip
+from io import BufferedWriter, BytesIO
+from typing import List, Optional, Sequence, Tuple, Union
 
-import dask.dataframe as dd
-import pandas as pd  # type: ignore
+import pandas as pd
 
 from countess import VERSION
+from countess.core.logger import Logger
 from countess.core.parameters import (
     ArrayParam,
     BooleanParam,
@@ -15,47 +16,31 @@ from countess.core.parameters import (
     MultiParam,
     StringParam,
 )
-from countess.core.plugins import DaskBasePlugin, DaskInputPlugin
+from countess.core.plugins import PandasInputFilesPlugin, PandasOutputPlugin
+from countess.utils.pandas import flatten_columns
 
-# XXX it would be better to do the same this Regex Tool does and get the user to assign
-# data types to each column
+CSV_FILE_TYPES: Sequence[Tuple[str, Union[str, List[str]]]] = [
+    ("CSV", [".csv", ".csv.gz"]),
+    ("TSV", [".tsv", ".tsv.gz"]),
+    ("TXT", [".txt", ".txt.gz"]),
+]
 
 
-def maybe_number(x):
-    """CSV is never clear on if something is actually a number so ... try it I guess ..."""
-    try:
-        return int(x)
-    except ValueError:
-        pass
-
-    try:
-        return float(x)
-    except ValueError:
-        pass
-
-    return x
-
-
-def clean_row(row):
-    return [maybe_number(x) for x in row]
-
-
-class LoadCsvPlugin(DaskInputPlugin):
+class LoadCsvPlugin(PandasInputFilesPlugin):
     """Load CSV files"""
 
     name = "CSV Load"
     description = "Loads data from CSV or similar delimited text files and assigns types to columns"
-    link = "https://countess-project.github.io/CountESS/plugins/#csv-reader"
+    link = "https://countess-project.github.io/CountESS/included-plugins/#csv-reader"
     version = VERSION
-
-    file_types = [("CSV", "*.csv"), ("TSV", "*.tsv"), ("TXT", "*.txt")]
+    file_types = CSV_FILE_TYPES
 
     parameters = {
         "delimiter": ChoiceParam("Delimiter", ",", choices=[",", ";", "TAB", "|", "WHITESPACE"]),
-        "quoting": ChoiceParam(
-            "Quoting", "None", choices=["None", "Double-Quote", "Quote with Escape"]
-        ),
+        "quoting": ChoiceParam("Quoting", "None", choices=["None", "Double-Quote", "Quote with Escape"]),
         "comment": ChoiceParam("Comment", "None", choices=["None", "#", ";"]),
+        "header": BooleanParam("CSV file has header row?", True),
+        "filename_column": StringParam("Filename Column", ""),
         "columns": ArrayParam(
             "Columns",
             MultiParam(
@@ -67,8 +52,6 @@ class LoadCsvPlugin(DaskInputPlugin):
                 },
             ),
         ),
-        "header": BooleanParam("CSV file has header row?", True),
-        "filename_column": StringParam("Filename Column", ""),
     }
 
     def read_file_to_dataframe(self, file_params, logger, row_limit=None):
@@ -80,6 +63,8 @@ class LoadCsvPlugin(DaskInputPlugin):
         if row_limit is not None:
             options["nrows"] = row_limit
 
+        index_col_numbers = []
+
         if len(self.parameters["columns"]):
             options["names"] = []
             options["dtype"] = {}
@@ -87,9 +72,11 @@ class LoadCsvPlugin(DaskInputPlugin):
 
             for n, pp in enumerate(self.parameters["columns"]):
                 options["names"].append(pp["name"].value or f"column_{n}")
-                if not pp["type"].is_none():
-                    options["dtype"][n] = pp["type"].get_selected_type()
+                if pp["type"].is_not_none():
+                    if pp["index"].value:
+                        index_col_numbers.append(len(options["usecols"]))
                     options["usecols"].append(n)
+                    options["dtype"][n] = pp["type"].get_selected_type()
 
         delimiter = self.parameters["delimiter"].value
         if delimiter == "TAB":
@@ -114,7 +101,6 @@ class LoadCsvPlugin(DaskInputPlugin):
         if comment != "None":
             options["comment"] = comment
 
-        # XXX dd.read_csv().set_index() is very very slow!
         # XXX pd.read_csv(index_col=) is half the speed of pd.read_csv().set_index()
 
         df = pd.read_csv(filename, **options)
@@ -126,27 +112,24 @@ class LoadCsvPlugin(DaskInputPlugin):
             for n, col in enumerate(df.columns):
                 if not self.parameters["columns"][n]["name"].value:
                     self.parameters["columns"][n]["name"].value = str(col)
+                    self.parameters["columns"][n]["type"].value = "string"
 
         filename_column = self.parameters["filename_column"].value
         if filename_column:
             df[filename_column] = filename
 
-        index_cols = [
-            df.columns[n] for n, pp in enumerate(self.parameters["columns"]) if pp["index"].value
-        ]
-        if index_cols:
-            df = df.set_index(index_cols)
+        if index_col_numbers:
+            df = df.set_index([df.columns[n] for n in index_col_numbers])
 
         return df
 
 
-class SaveCsvPlugin(DaskBasePlugin):
+class SaveCsvPlugin(PandasOutputPlugin):
     name = "CSV Save"
     description = "Save data as CSV or similar delimited text files"
-    link = "https://countess-project.github.io/CountESS/plugins/#csv-writer"
+    link = "https://countess-project.github.io/CountESS/included-plugins/#csv-writer"
     version = VERSION
-
-    file_types = [("CSV", "*.csv"), ("TSV", "*.tsv"), ("TXT", "*.txt")]
+    file_types = CSV_FILE_TYPES
 
     parameters = {
         "header": BooleanParam("CSV header row?", True),
@@ -155,38 +138,56 @@ class SaveCsvPlugin(DaskBasePlugin):
         "quoting": BooleanParam("Quote all Strings", False),
     }
 
-    def run(
-        self,
-        data: Any,
-        logger,
-        row_limit: Optional[int] = None,
-    ):
-        assert isinstance(self.parameters["filename"], StringParam)
+    filehandle: Optional[Union[BufferedWriter, BytesIO]] = None
+    csv_columns = None
 
-        filename = self.parameters["filename"].value
-        sep = self.parameters["delimiter"].value
-        if sep == "TAB":
-            sep = "\t"
-        elif sep == "SPACE":
-            sep = " "
-
-        options = {
-            "sep": sep,
-            "quoting": csv.QUOTE_NONNUMERIC
-            if self.parameters["quoting"].value
-            else csv.QUOTE_MINIMAL,
-        }
+    SEPARATORS = {",": ",", ";": ";", "SPACE": " ", "TAB": "\t"}
+    QUOTING = {False: csv.QUOTE_MINIMAL, True: csv.QUOTE_NONNUMERIC}
 
+    def prepare(self, sources: list[str], row_limit: Optional[int] = None):
         if row_limit is None:
-            if isinstance(data, dd.DataFrame):
-                data.to_csv(filename, single_file=True, compute=True, **options)
+            filename = self.parameters["filename"].value
+            if filename.endswith(".gz"):
+                self.filehandle = gzip.open(filename, "wb")
             else:
-                data.to_csv(filename, **options)
-            return None
+                self.filehandle = open(filename, "wb")
         else:
-            if isinstance(data, dd.DataFrame):
-                data = data.compute()
+            self.filehandle = BytesIO()
 
-            buf = StringIO()
-            data.to_csv(buf, **options)
-            return buf.getvalue()
+        self.csv_columns = None
+
+    def process(self, data: pd.DataFrame, source: str, logger: Logger):
+        # reset indexes so we can treat all columns equally.
+        # if there's just a nameless index then we don't care about it, drop it.
+        drop_index = data.index.name is None and data.index.names[0] is None
+        dataframe = flatten_columns(data.reset_index(drop=drop_index))
+
+        # if this is our first dataframe to write then decide whether to
+        # include the header or not.
+        if self.csv_columns is None:
+            self.csv_columns = list(dataframe.columns)
+            emit_header = bool(self.parameters["header"].value)
+        else:
+            # add in any columns we haven't seen yet in previous dataframes.
+            for c in dataframe.columns:
+                if c not in self.csv_columns:
+                    self.csv_columns.append(c)
+                    logger.warning(f"Added CSV Column {repr(c)} with no header")
+            # fill in blanks for any columns which are in previous dataframes but not
+            # in this one.
+            dataframe = dataframe.assign(**{c: None for c in self.csv_columns if c not in dataframe.columns})
+            emit_header = False
+
+        dataframe.to_csv(
+            self.filehandle,
+            header=emit_header,
+            columns=self.csv_columns,
+            index=False,
+            sep=self.SEPARATORS[self.parameters["delimiter"].value],
+            quoting=self.QUOTING[self.parameters["quoting"].value],
+        )  # type: ignore [call-overload]
+        return []
+
+    def finalize(self, logger: Logger):
+        if isinstance(self.filehandle, BytesIO):
+            yield self.filehandle.getvalue().decode("utf-8")
