@@ -1,28 +1,20 @@
 import time
+
 import pytest
 
-from countess.gui.main import make_root, MainWindow
 from countess.core.config import write_config
+from countess.gui.main import MainWindow, make_root
 
+
+@pytest.mark.gui
 def test_open_nodes():
-
     root = make_root()
 
     mw = MainWindow(root, "tests/simple.ini")
 
     root.update()
 
-    nodes = list(mw.graph_wrapper.graph.traverse_nodes())
-
-    last_node = nodes[-1]
-
-    mw.graph_wrapper.on_mousedown(last_node, None)
-
-    root.update()
-
-    assert last_node.result
-    
-    for node in nodes:
+    for node in mw.graph_wrapper.graph.traverse_nodes():
         mw.graph_wrapper.on_mousedown(node, None)
         root.update()
         time.sleep(0.1)
